# -*- coding: utf-8 -*-

import click
import os
import pandas as pd
import numpy as np
from boba.bobarun import BobaRun
from bobaserver import app, socketio, scheduler
from .util import read_json, write_json, read_key_safe, print_fail
import bobaserver.common as common


def check_path(p, more=''):
    """Check if the path exists"""
    if not os.path.exists(p):
        msg = 'Error: {} does not exist.'.format(p)
        print_help(msg + more)


def check_required_field(obj, key, prefix=''):
    """Check if a required field is in obj."""
    if key not in obj:
        err = 'Error: cannot find required field "{}" in {}'.format(key, obj)
        print_help(prefix + err)

def print_help(err=''):
    """Show help message and exit."""
    ctx = click.get_current_context()
    click.echo(ctx.get_help())

    if err:
        click.secho('\n' + err, fg='red')
    ctx.exit()


def read_meta():
    """ Read overview.json, verify, and store the meta data. """
    fn = os.path.join(app.data_folder, 'overview.json')
    err, res = read_json(fn)
    if (err):
        print_help(err['message'])

    # check summary.csv
    fn = os.path.join(app.data_folder, 'summary.csv')
    check_path(fn)

    # check file definition
    vis = read_key_safe(res, ['visualizer'], {})
    fs = read_key_safe(vis, ['files'], [])
    lookup = {}
    prefix = 'In parsing visualizer.files in overview.json:\n'
    for f in fs:
        check_required_field(f, 'id', prefix)
        check_required_field(f, 'path', prefix)
        f['multi'] = read_key_safe(f, ['multi'], False)
        lookup[f['id']] = f

    # read schema and join file
    schema = read_key_safe(vis, ['schema'], {})
    prefix = 'In parsing visualizer.schema in overview.json:\n'
    check_required_field(schema, 'point_estimate', prefix)
    for key in schema:
        s = schema[key]
        check_required_field(s, 'file', prefix)
        # check_required_field(s, 'field', prefix)
        # todo: verify if file is valid CSV and if field exist in file
        fid = s['file']
        if fid not in lookup:
            msg = 'In parsing visualizer.schema in overview.json:\n'
            msg += '{}\n'.format(s)
            msg += 'Error: file id "{}" is not defined.'.format(fid)
            print_help(msg)
        s['file'] = lookup[fid]['path']
        s['multi'] = lookup[fid]['multi']
        s['name'] = key

    # check sensitivity flag
    sen = read_key_safe(vis, ['sensitivity'], 'ad')
    if sen not in ('f', 'ks', 'ad'):
        msg = f'Invalid sensitivity flag "{sen}". Available values:\n'
        msg += ' - "f": algorithm based on the F-test\n'
        msg += ' - "ks": algorithm based on Kolmogorov–Smirnov statistic'
        msg += ' - "ad": k-samples Anderson-Darling test'
        print_help(msg)

    # store meta data
    app.schema = schema
    app.summary = common.read_summary()
    app.decisions = read_key_safe(res, ['decisions'], {})
    app.visualizer = {
        "sensitivity": sen,
        "labels": read_key_safe(vis, ['labels'], {}),
        "graph": read_key_safe(res, ['graph'], {})
    }


def check_result_files ():
    """ check if result files exists """
    fn = os.path.join(app.data_folder, 'overview.json')
    err, res = read_json(fn)
    vis = read_key_safe(res, ['visualizer'], {})
    fs = read_key_safe(vis, ['files'], [])

    prefix = 'In parsing visualizer.files in overview.json:\n'
    for f in fs:
        multi = read_key_safe(f, ['multi'], False)
        if not multi:
            check_path(os.path.join(app.data_folder, f['path']))


@click.command()
@click.option('--in', '-i', 'input', default='.', show_default=True,
              help='Path to the input directory')
@click.option('--port', default=8080, show_default=True,
              help='The port to bind the server to')
@click.option('--host', default='0.0.0.0', show_default=True,
              help='The interface to bind the server to')
@click.option('--monitor', is_flag=True, help='Allow boba monitor')
@click.version_option()
def main(input, port, host, monitor):
    check_path(input)
    app.data_folder = os.path.realpath(input)

    read_meta()
    app.bobarun = BobaRun(app.data_folder)
    if not monitor:
        check_result_files()

        # compute sensitivity and write scores to file
        app.sensitivity = common.cal_sensitivity()
        d = {'method': app.visualizer['sensitivity'], 'scores': app.sensitivity}
        write_json(d, os.path.join(input, 'sensitivity.json'), nice=True)

    # print starting message
    s_host = '127.0.0.1' if host == '0.0.0.0' else host
    msg = """\033[92m
    Server started!
    Navigate to http://{0}:{1}/ in your browser
    Press CTRL+C to stop\033[0m""".format(s_host, port)
    print(msg)

    # start server
    scheduler.start()
    if monitor:
        socketio.run(app, host= host, port=f'{port}')
    else:
        app.run(host= host, port=f'{port}')


if __name__ == '__main__':
    main()



from scipy import stats
import numpy as np
import warnings


def sensitivity_ad (df, dec, options, col):
  """ use the k-sample Anderson-Darling test to compute sensitivity """
  if len(options) < 2:
    return 0, 1

  groups = []
  for opt in options:
      groups.append(df.loc[df[dec] == opt][col].tolist())

  # groupby is incorrect because a decision can be omitted due to dependency
  # the decision column would have empty value in summary.csv
  # groups = df.groupby(dec)[col].apply(list).tolist()

  with warnings.catch_warnings():
    # suppress the warning "p-value capped: true value larger than 0.25"
    warnings.simplefilter('ignore')

    # run the test
    ad = stats.anderson_ksamp(groups)

  # normalized test statistics and p-value
  return ad.statistic, ad.significance_level


def ad_wrapper (df, dec, col):
  """
  Run AD test for a given decision, while checking for minimum sample
  size requirements. Returns NaN if the check or the AD test fails.

  Returns: (test statistics, p-value)
  """
  # each option should have some samples for the k-samples AD test to work
  min_group_size = 3

  # ensure that each level has at least n samples
  groups = df.groupby(dec).count()
  n_pass = groups[groups[col] >= min_group_size].shape[0]
  if n_pass < groups.shape[0]:
    return np.nan, np.nan

  # we are using the options in the current df, ignoring missing levels
  options = df.groupby(dec).groups.keys()
  try:
    s, p = sensitivity_ad(df, dec, options, col)
    return s, p
  except (ValueError, IndexError):
    return np.nan, np.nan


def sensitivity_ks (df, dec, options, col):
  """ compute Kolmogorov-Smirnov statistic """
  if len(options) < 2:
    return 0

  groups = []
  for opt in options:
      groups.append(df.loc[df[dec] == opt][col].to_numpy())

  kss = []
  for i in range(len(groups)):
      for j in range(i + 1, len(groups)):
          ks = stats.ks_2samp(groups[i], groups[j])
          kss.append(ks.statistic)  # ks.pvalue gives p-value

  # median KS stat
  return np.median(kss)


def sensitivity_f (df, dec, options, col):
  """ Compute one-way F-test to estimate decision sensitivity """
  if len(options) < 2:
    return 0

  x_mean = df[col].mean()

  groups = []
  for opt in options:
      groups.append(df.loc[df[dec] == opt][['uid', col]])

  # ms between
  ms_b = 0
  for g in groups:
      ms_b += len(g) * (g[col].mean() - x_mean)**2
  ms_b /= len(groups) - 1

  # ms within
  ms_w = 0
  for g in groups:
      g_mean = g[col].mean()
      ms_w += sum((g[col] - g_mean)**2)
  ms_w /= len(df) - len(groups)

  return np.sqrt(ms_b / ms_w)



# -*- coding: utf-8 -*-
import os
import csv
import json
import pandas as pd
import numpy as np


class Colors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def print_fail(msg):
    print(Colors.FAIL + msg + Colors.ENDC)


def print_warn(msg):
    print(Colors.OKBLUE + msg + Colors.ENDC)


def check_path(fn):
    """ Check if path exists """
    if not os.path.exists(fn):
        msg = 'Error: {} does not exist.'.format(fn)
        return {'status': 'fail', 'message': msg}


def read_csv(fn, row_start=1):
    """ Read csv with path check, discarding (optionally) the first row """
    err = check_path(fn)
    if err:
        return err, None

    res = []
    with open(fn, 'r', newline='') as f:
        reader = csv.reader(f, delimiter=',')
        for row in reader:
            res.append(row)
    return err, res[row_start:]


def read_json(fn):
    """ Read a JSON file with path check"""
    err = check_path(fn)
    if err:
        return err, None

    with open(fn, 'rb') as f:
        try:
            res = json.load(f)
            return None, res
        except json.JSONDecodeError:
            msg = 'Cannot parse the JSON file {}'.format(fn)
            err = {'status': 'fail', 'message': msg}
            return err, None

def write_json(data, fn, nice=False):
    param = {'indent': 4, 'sort_keys': True} if nice else {}
    with open(fn, 'w', encoding='utf-8') as f:
        json.dump(data, f, **param)


def read_key_safe(obj, keys, default):
    """ Recursively check if key is in obj and return the value.
    Otherwise, return default. """
    for key in keys:
        if key in obj:
            obj = obj[key]
        else:
            return default
    
    return obj

def group_by(lst, func):
    res = {}
    for item in lst:
        k = func(item)
        if k in res:
            res[k].append(item)
        else:
            res[k] = [item]
    return res

def remove_na (df, col, dtype=str):
    """ convert a column in a dataframe to a data type, and remove any rows
    with Inf or NA values in this column"""

    if dtype != str:
        dc = 'float' if dtype == float else 'integer'
        df[col] = pd.to_numeric(df[col], errors='coerce', downcast=dc)

    # remove Inf and NA
    df = df.replace([np.inf, -np.inf], np.nan)
    df = df.dropna(subset=[col])

    return df



import numpy as np
from scipy import stats
from time import perf_counter

class bootstrap():
  """ 
  This class implements the bootstrap procedure and offers 3 type of CIs:
  basic, percentile, and bias-corrected and accelerated (BCa).
  """

  def __init__(self, func, ci_type='percentile', n=200, verbose=False):
    """
    Parameters:
     - n: bootstrap how many times
     - func: function to compute the statistic of interest
     - ci_type: one of ['basic', 'percentile', 'bca', 'bc']
     - verbose: if true, print elapsed time
    """
    self.n = n
    self.stat = func
    self.ci_type = ci_type
    self.verbose = verbose


  def fit(self, data, *args, **kwargs):
    self.fit_with_probability(data, None, *args, **kwargs)


  def fit_with_probability(self, data, p, *args, **kwargs):
    """
    Parameters:
     - data: the sample array to perform bootstrap on
     - p: probilities associated with each entry in data. If unspecified, draw
       uniformly when resampling.
    """
    time_start = perf_counter()

    # sample statistics
    self.sample_stat = self.stat(data, *args, **kwargs)

    # fit bootstrap
    self.bootstrap_stats = []
    for i in range(self.n):
      d = np.random.choice(data, size=len(data), replace=True, p=p)
      self.bootstrap_stats.append(self.stat(d, *args, **kwargs))

    # jackknife
    self.jack_stats = []
    if self.ci_type == 'bca':
      for i in range(len(data)):
        d = np.delete(data, i)
        self.jack_stats.append(self.stat(d, *args, **kwargs))

    # elapsed time
    if self.verbose:
      print(f'Bootstrap time: {perf_counter() - time_start} seconds')


  def get_ci(self, alpha=0.05):
    """ Get the 100(1 - alpha)% confidence interval"""
    # drop NaN in the bootstrap statistics array
    arr = self._handle_null(self.bootstrap_stats)
    if not len(arr):
      return np.nan, np.nan

    if self.ci_type == 'basic':
      return self._get_ci_basic(arr, alpha)
    elif self.ci_type == 'bca':
      return self._get_ci_bca(arr, alpha)
    elif self.ci_type == 'bc':
      return self._get_ci_bc(arr, alpha)
    else:
      return self._get_ci_percentile(arr, alpha)


  def _get_ci_bc(self, arr, alpha):
    """ BCa bootstrap CI with the acceleration term set to 0 """
    # bias-correction factor
    z0 = stats.norm.ppf(np.mean(np.asarray(arr) < self.sample_stat))

    # the CI
    ql, qu = stats.norm.ppf(alpha / 2), stats.norm.ppf(1 - alpha / 2)
    a1 = stats.norm.cdf(z0 * 2 + ql)
    a2 = stats.norm.cdf(z0 * 2 + qu)
    return np.quantile(arr, a1), np.quantile(arr, a2)


  def _get_ci_bca(self, arr, alpha):
    """ Bias corrected and accelerated bootstrap CI """
    # bias-correction factor
    z0 = stats.norm.ppf(np.mean(np.asarray(arr) < self.sample_stat))
    z0 = np.sign(z0) * 5 if np.isinf(z0) else z0  # handle infinity

    # acceleration factor
    jack = np.asarray(self._handle_null(self.jack_stats))
    if not len(jack):
      a_hat = 0
    else:
      nom = np.sum((jack.mean() - jack)**3)
      denom = 6 * np.sum((jack.mean() - jack)**2)**1.5
      a_hat = nom / denom

    # the CI
    ql, qu = stats.norm.ppf(alpha / 2), stats.norm.ppf(1 - alpha / 2)
    a1 = stats.norm.cdf(z0 + (z0 + ql) / (1 - a_hat * (z0 + ql)))
    a2 = stats.norm.cdf(z0 + (z0 + qu) / (1 - a_hat * (z0 + qu)))
    return np.quantile(arr, a1), np.quantile(arr, a2)


  def _get_ci_basic(self, arr, alpha):
    # basic CI
    lower = 2 * self.sample_stat - np.quantile(arr, 1 - alpha / 2)
    upper = 2 * self.sample_stat -np.quantile(arr, alpha / 2)
    return lower, upper


  def _get_ci_percentile(self, arr, alpha):
    # percentile based CI
    lower = np.quantile(arr, alpha / 2)
    upper = np.quantile(arr, 1 - alpha / 2)
    return lower, upper


  def _handle_null (self, arr):
    # drop NaN in the bootstrap/jackknife statistics array
    return [t for t in arr if not np.isnan(t)]



# routes related to the boba run monitor

import os
import time
import pandas as pd
import numpy as np
from flask import jsonify, request
from .util import read_csv, read_json, write_json
from bobaserver import app, socketio, scheduler
from bobaserver.bobastats import sampling, sensitivity
import bobaserver.common as common


class BobaWatcher:
  # static attributes
  header_outcome = ['n_samples', 'mean', 'lower', 'upper']

  def __init__(self, order, weights=None):
    self.start_time = None
    self.prev_time = 0  # for resume

    # sampling order and weights
    self.order = [uid - 1 for uid in order]  # convert to 0-indexed
    self.weights = weights

    # results
    self.last_merge_index = 0
    self.outcomes = []
    self.decision_scores = []


  @staticmethod
  def get_fn_outcome():
    return os.path.join(app.bobarun.dir_log, 'outcomes.csv')
  @staticmethod
  def get_fn_save():
    return os.path.join(app.bobarun.dir_log, 'execution_plan.json')
  @staticmethod
  def get_fn_sensitivity():
    return os.path.join(app.bobarun.dir_log, 'sensitivity.csv')
  @staticmethod
  def get_header_sensitivity():
    return ['n_samples', 'type'] + common.get_decision_list()


  def _append_csv(self, fn, header, data):
    # append to the csv if it exists, or create one
    if os.path.exists(fn):
      f = open(fn, 'a')
    else:
      f = open(fn, 'w')
      f.write(','.join(header) + '\n')
    for r in data:
      f.write(','.join([str(i) for i in r]) + '\n')
    f.close()


  def _impute_null_CI(self, data, previous, col=0):
    # impute NaN in CIs, assuming data is a 2D list [..., mean, lower, upper]
    # where col is the column index of mean. Modify data in place.
    for i, d in enumerate(data):
      for j in [col + 1, col + 2]:
        if np.isnan(d[j]):
          d[j] = data[i - 1][j] if i > 0 else (previous[-1][j] \
            if len(previous) else d[col])


  def _NaN_to_string(self, arr):
    # convert NaN to string 'nan'. It will happen in place in the input arr.
    for i in range(len(arr)):
      arr[i] = ['nan' if isinstance(r, float) and np.isnan(r) else r \
        for r in arr[i]]


  def _compute_dec_CI(self, df, col, indices, dec_list, i):
    """ Compute bootstrap CI of decision sensitivity """
    res = sampling.bootstrap_sensitivity(df, col, indices, dec_list)
    out = [[i, c] + res[f'score_{c}'].tolist() for c in ['lower', 'upper']]

    # convert NaN to string
    self._NaN_to_string(out)
    self.decision_scores += out

    # write results to disk
    self._append_csv(BobaWatcher.get_fn_sensitivity(),
      BobaWatcher.get_header_sensitivity(), out)

    # send to client
    socketio.emit('update-sensitivity', {'data': self.decision_scores,
      'header': BobaWatcher.get_header_sensitivity()})


  def update_outcome(self, done):
    step = min(5, max(1, int(app.bobarun.size / 50)))
    if len(done) - self.last_merge_index <= step:
      return

    # merge result file
    app.bobarun.run_after_execute()
    df = common.read_results('point_estimate', float)
    df = pd.merge(app.summary, df, on='uid', how='left')
    col = common.get_field_name('point_estimate')
    dec_list = common.get_decision_list()

    # compute results since the last index
    start = (int(self.last_merge_index / step) + 1) * step
    self.last_merge_index = len(done) - 1
    res = []
    sen = []
    indices = None
    for i in range(start, len(done), step):
      indices = self.order[:i+1]

      # outcome mean
      out = sampling.bootstrap_outcome(df, col, indices, self.weights)
      res.append([i] + out)

      # decision sensitivity, without CI
      # FIXME: hard coded for AD test
      ad = [sensitivity.ad_wrapper(df.iloc[indices], dec, col) \
        for dec in dec_list]
      sen.append([i, 'score'] + [s[0] for s in ad])
      sen.append([i, 'p'] + [s[1] for s in ad])

    # schedule a job to compute the decision CI, for the last index
    if (not scheduler.get_job('compute_CI')) and (indices is not None):
      scheduler.add_job(self._compute_dec_CI, id='compute_CI', 
        args=[df, col, indices, dec_list, i])

    # impute null in CI and remove null in mean
    self._impute_null_CI(res, self.outcomes, 1)
    res = [r for r in res if not np.isnan(r[1])]
    self.outcomes += res

    # convert NaN in decision sensitivity to string 'nan'
    self._NaN_to_string(sen)
    self.decision_scores += sen

    # write results to disk
    self._append_csv(BobaWatcher.get_fn_outcome(), self.header_outcome, res)
    self._append_csv(BobaWatcher.get_fn_sensitivity(),
      BobaWatcher.get_header_sensitivity(), sen)

    # send to client
    socketio.emit('update-outcome', {'data': self.outcomes, 
      'header': self.header_outcome})
    socketio.emit('update-sensitivity', {'data': self.decision_scores,
      'header': BobaWatcher.get_header_sensitivity()})


  def check_progress(self):
    # remove self from scheduled jobs if boba run has finished
    if not app.bobarun.is_running():
      scheduler.remove_job('watcher')
    print('check progress')

    # estimate remaining time
    logs = app.bobarun.exit_code
    done = max(1, len(logs)) # avoid division by 0
    elapsed = self.get_elapsed()
    remain = app.bobarun.size - done
    remain = int(elapsed * remain / done)

    # schedule jobs to compute results
    if not scheduler.get_job('update_outcome'):
      scheduler.add_job(self.update_outcome, args=[logs], id='update_outcome')

    res = {'status': 'success',
      'logs': logs,
      'time_left': remain,
      'is_running': app.bobarun.is_running()}

    socketio.emit('update', res)


  def stop(self):
    # stop timer
    t = 0 if self.start_time is None else time.time() - self.start_time
    self.prev_time += t
    self.start_time = None


  def start(self):
    # start timer and add job
    self.start_time = time.time()
    scheduler.add_job(self.check_progress, 'interval', seconds=5,
      id='watcher', replace_existing=True)


  def get_elapsed(self):
    t = 0 if self.start_time is None else time.time() - self.start_time
    return t + self.prev_time


  def save_to_file(self):
    # save data to file, so it is possible to resume later
    data = {'order': list(self.order), 'elapsed': self.get_elapsed()}
    if self.weights is not None:
      data['weights'] = list(self.weights)
    write_json(data, self.get_fn_save())


  def init_from_file(self):
    # resume from the previous save
    err, data = read_json(self.get_fn_save())
    if not err:
      self.order = data['order']
      self.weights = np.asarray(data['weights']) if 'weights' in data else None
      self.prev_time = data['elapsed']

    # read outcome and sensitivity progress
    fn = BobaWatcher.get_fn_outcome()
    if os.path.exists(fn):
      df = pd.read_csv(fn)
      self.last_merge_index = df['n_samples'].max()
      self.outcomes = df.values.tolist()
    fn = BobaWatcher.get_fn_sensitivity()
    if os.path.exists(fn):
      # convert NaN to string 'nan'; client needs to convert it back to js NaN
      df = pd.read_csv(fn).fillna('nan')
      self.decision_scores = df.values.tolist()


def check_stopped():
  # after client issued stop command, check if boba has indeed stopped
  if not app.bobarun.is_running():
    scheduler.remove_job('check_stopped')
    socketio.emit('stopped')


def merge_error ():
  """ Merge the error logs into errors.csv """
  fn = os.path.join(app.bobarun.dir_log, 'errors.csv')
  logs = []
  merged = []
  df = None

  # exit code
  if os.path.exists(app.bobarun.file_log):
    status = pd.read_csv(app.bobarun.file_log, index_col='uid')
    logs = status.index.tolist()

  # previous merged error
  if os.path.exists(fn):
    df = pd.read_csv(fn, na_filter=False)
    merged = df['uid'].tolist()

  # these are the new logs
  remain = set(logs).difference(set(merged))
  res = []
  for f in os.listdir(app.bobarun.dir_log):
    if f.startswith('error') and f.endswith('txt'):
      uid = int(os.path.splitext(f)[0].split('_')[1])
      if uid in remain:
        with open(os.path.join(app.bobarun.dir_log, f), 'r') as fo:
          data = fo.read()
          code = status.loc[uid]['exit_code']
          res.append([uid, code, data])

  # cluster errors into groups
  res = pd.DataFrame(res, columns=['uid', 'exit_code', 'message'])
  res = common.cluster_error(res)

  # save file and return
  df = res if df is None else pd.concat([df, res], ignore_index=True)
  df.to_csv(fn, index=False)
  return df, status


# entry (already defined in routes)
# @app.route('/')
# def index():
#   return app.send_static_file('index.html')


@app.route('/api/monitor/start_runtime', methods=['POST'])
def start_runtime():
  # compute sampling order and weights
  # TODO: allow users to specify the sampling method
  df = common.get_decision_df()
  order, weights = sampling.round_robin(df, df.shape[0])

  # order is not a list of uid, but indices into the summary table
  # lookup again to get the actual uid
  order = common.read_summary().iloc[order]['uid'].tolist()
  # compute likelihood ratio
  if weights is not None:
    weights = 1 / (weights * df.shape[0])

  fresh = not app.bobarun.is_running()

  # if last run is completed, instead of stopped, the job will still remain
  # in the scheduler. We need to remove the job to start again.
  if fresh and scheduler.get_job('bobarun'):
    scheduler.remove_job('bobarun')

  if fresh:
    # periodic check for progress
    app.bobawatcher = BobaWatcher(order, weights)
    app.bobawatcher.start()

  # set batch size to 1 so the log would be updated more frequently
  app.bobarun.batch_size = 1

  # the scheduler will ensure that we have only 1 running instance
  scheduler.add_job(app.bobarun.run_multiverse, args=[order], id='bobarun')

  return jsonify({'status': 'success'}), 200


@app.route('/api/monitor/resume_runtime', methods=['POST'])
def resume_runtime():
  # return error if bobarun is still running
  if app.bobarun.is_running():
    return jsonify({'status': 'fail', 'message': 'Boba is still running'}), 200

  # initialize boba watcher if this is a new instance
  if not hasattr(app, 'bobawatcher'):
    app.bobawatcher = BobaWatcher([])
    app.bobawatcher.init_from_file()

  # check if boba watcher is valid
  if len(app.bobawatcher.order) < 1:
    return jsonify({'status': 'fail',
      'message': 'Cannot find the previous execution plan'}), 200

  # convert order to 1-indexed
  order = [uid + 1 for uid in app.bobawatcher.order]

  # start runtime, as we do in start_runtime
  app.bobawatcher.start()
  app.bobarun.batch_size = 1
  scheduler.add_job(app.bobarun.resume_multiverse, args=[order], id='bobarun')

  return jsonify({'status': 'success'}), 200


@app.route('/api/monitor/stop_runtime', methods=['POST'])
def stop_runtime():
  # ask boba run to stop, but post_exe.sh will still run
  app.bobarun.stop()

  # save the plan, so we can resume later
  if hasattr(app, 'bobawatcher'):
    app.bobawatcher.stop()
    app.bobawatcher.save_to_file()

  # periodically check if boba run has indeed stopped
  scheduler.add_job(check_stopped, 'interval', seconds=1, id='check_stopped')

  return jsonify({'status': 'success'}), 200


@app.route('/api/monitor/get_snapshot', methods=['POST'])
def get_snapshot():
  res = {'status': 'success',
    'results': {'data': [], 'header': []},
    'errors': {'data': [], 'header': []}}

  if not os.path.exists(app.bobarun.dir_log):
    return jsonify(res), 200

  # error messages
  err_msg, exit_code = merge_error()
  res['errors']['data'] = err_msg.values.tolist()
  res['errors']['header'] = err_msg.columns.tolist()

  # perform merge because the last merge may be stale
  app.bobarun.run_after_execute()

  # read results and keep NA
  fields = ['point_estimate', 'p_value', 'fit']
  df = common.read_results_batch(fields)
  df = pd.merge(exit_code, df, on='uid', how='left').fillna('nan')
  res['results']['data'] = df.values.tolist()
  res['results']['header'] = df.columns.tolist()

  return jsonify(res), 200


@app.route('/api/monitor/inquire_progress', methods=['POST'])
def inquire_progress():
  res = {'status': 'success',
    'logs': [],
    'outcome': {'data': [], 'header': BobaWatcher.header_outcome},
    'decision_scores': {'data': [], 'header': []},
    'size': app.bobarun.size,
    'is_running': app.bobarun.is_running()}

  # exit code
  if os.path.exists(app.bobarun.file_log):
    err, t = read_csv(app.bobarun.file_log)
    res['logs'] = t

  # outcome CI time series
  if not hasattr(app, 'bobawatcher'):
    app.bobawatcher = BobaWatcher([])
    app.bobawatcher.init_from_file()
  res['outcome']['data'] = app.bobawatcher.outcomes
  res['decision_scores']['data'] = app.bobawatcher.decision_scores
  res['decision_scores']['header'] = app.bobawatcher.get_header_sensitivity()

  # for debugging
  # logs = [int(r[0]) for r in res['logs']]
  # BobaWatcher(logs).update_outcome(logs)

  return jsonify(res), 200



import numpy as np
import pandas as pd
import itertools
from sklearn.linear_model import LinearRegression
from .bootstrap import bootstrap
from .sensitivity import ad_wrapper


def round_robin (df, n=50):
  """
  Round robin baseline (similar to stratified sampling).
  """
  df_snapshot = df
  df = df.copy()
  decs = list(df.columns)

  # weights
  weights = round_robin_weights(df)

  indices = []
  while len(indices) < n and df.shape[0] > 0:
    # in each round, go over each decision
    for dec in decs:
      gp = df.groupby(dec)
      # go over each option
      for opt in gp.groups.keys():
        # draw an index from all rows of the option
        row_ids = gp.groups[opt]
        drawn_id = row_ids[np.random.choice(len(row_ids))]
        indices.append(drawn_id)

        # remove from df so we are sampling without replacement
        df = df.drop(index=drawn_id)

  # convert to position-based index (iloc)
  indices = df_snapshot.index.get_indexer(indices)
  return indices, weights


def uniform (df, n=50):
  """
  Sample universes uniformly and return the sketching indices
  """
  indices = np.random.choice(df.shape[0], n, False)
  return indices, None


def sketching (df, n=50, interact=False):
  """
  Use a random sampling-based sketching algorithm without rescaling; 
  It is algorithm 2 on page 8. Return the normalized leverage scores
  and the sketching indices.
  """
  # one hot encoding
  X = one_hot_encode(df, interact=interact)

  # calculate leverage scores
  U, s, V = np.linalg.svd(X, full_matrices=False)
  l = np.dot(U, np.transpose(U)).diagonal()

  # get distribution and sample
  p = X.shape[1]
  dist = l / p
  indices = np.random.choice(X.shape[0], n, False, dist)

  return indices, dist


def round_robin_weights (df):
  """
  Compute the probability of drawing each universe in the first round of round
  robin. The probability is normalized (the sum of all universes is 1).

  Returns: a numpy array with the probability of drawing the universe where the
    index matches the index in the input df.
  """
  # a list of decisions
  decs = df.columns.tolist()
  df = df.copy().fillna('')
  df['dummy'] = 0

  # construct a lookup table for marginal probabilities
  lookup = {}
  for d in decs:
    gp = df.groupby(d)
    options = gp.groups.keys()
    lookup[d] = {opt: 1 / gp.count()['dummy'].loc[opt] for opt in options}

  # probability for a universe to be drawn in the first round
  def compute_prob (row):
    marginal = [lookup[d][row[d]] for d in decs]
    p = 0
    for i in range(1, len(decs)):
      v = np.sum([np.prod(c) for c in itertools.combinations(marginal, i)])
      p += v if i % 2 else (-v)
    return p

  weights = df.apply(compute_prob, axis=1).to_numpy()
  weights = weights / np.sum(weights) # normalize
  return weights


def get_outcome_mean (y, indices, weights=None, ignore_na=True):
  """
  Estimate outcome mean from sample.
    - weights: likelihood ratio f(x)/g(x) for importance sampling
  """
  arr = y[indices]
  if weights is not None:
    arr = weights[indices] * arr
  if ignore_na:
    arr = arr[~np.isnan(arr)]

  return np.mean(arr)


def bootstrap_outcome (df, COL, indices, weights=None, ignore_na=True):
  """
  Given a sample, compute the bootstrapped CI around outcome mean.

  Parameters:
   - df: multiverse dataframe
   - COL: the column in df
   - indices: sample index into the multiverse df
   - weights: importance sampling weights, if applicable
  """
  y = df[COL].to_numpy()
  mean = get_outcome_mean(y, indices, weights)

  # we will pass the index array to bootstrap, so here we adjust the func API
  stat = lambda idx, w: get_outcome_mean(y, idx, w)

  # bootstrap
  bs = bootstrap(stat, ci_type='percentile', n=200)
  bs.fit(indices, weights)  # sample uniformly, weighted mean
  lower, upper = bs.get_ci()

  return [mean, lower, upper]


def bootstrap_sensitivity (df, COL, indices, decs=None):
  """ Sensitivity and bootstrapped CI for all decisions. """
  if decs is None:
    # assuming all columns except "outcome" is a decision
    decs = list(df.columns)
    decs.remove(COL)

  # prep work
  out = []
  header = ['decision', 'score', 'p', 'score_lower', 'score_upper']
  # our bootstrap statisitc is the AD score of a decision
  stat = lambda idx, d: ad_wrapper(df.iloc[idx], d, COL)[0]

  # loop over all decisions
  for d in decs:
    # sample stats
    score, pval = ad_wrapper(df.iloc[indices], d, COL)
    row = [d, score, pval]

    # bootstrap
    if not np.isnan(score):
      bs = bootstrap(stat, ci_type='bc', n=200)
      bs.fit(indices, d)
      lower, upper = bs.get_ci()
      row += [lower, upper]

    # pad with NaN if we did not bootstrap
    row += [np.nan] * (len(header) - len(row))
    out.append(row)

  return pd.DataFrame(out, columns = header)


def one_hot_encode (df, interact=False):
  """
  Convert data into dummy coding for linear regression.
  Parameter:
   - interact: whether to add all possible two-way interactions
  """
  # a list of decisions
  decisions = df.columns.tolist()

  # helper function to create interaction terms
  REF_LEVEL = 'reference_level'
  ref = df.iloc[0]
  def add_interaction (row):
    # first manually replace reference level
    for dec in decisions:
      row[dec] = REF_LEVEL if row[dec] == ref[dec] else row[dec]

    # now add all two-way interactions
    for i in range(0, len(decisions)):
      for j in range(i + 1, len(decisions)):
        d1 = decisions[i]
        d2 = decisions[j]
        val = REF_LEVEL if (row[d1] == REF_LEVEL or row[d2] == REF_LEVEL) \
          else f'{row[d1]}_{row[d2]}'
        row[f'{d1}_{d2}'] = val

    return row

  # one-hot encoding
  if not interact:
    df = pd.get_dummies(df, columns = decisions, drop_first=True)
  else:
    df = df.apply(add_interaction, axis=1)
    cols = df.columns.tolist()
    df = pd.get_dummies(df, columns=cols)
    # now manually remove any columns with ref level
    cols = df.columns.tolist()
    cols = [c for c in cols if REF_LEVEL in c]
    df = df.drop(columns=cols, axis=1)

  # get X
  X = df.to_numpy()
  return X



import pandas as pd
import numpy as np
import os
import re
from bobaserver import app
from .bobastats import sensitivity
from .util import print_warn, remove_na, group_by


def get_decision_list ():
  # get a list of decision names
  return sorted([d['var'] for d in app.decisions])


def get_decision_df ():
  # get the summary.csv without any non-decision columns
  dec = [d['var'] for d in app.decisions]
  return read_summary()[dec]


def get_field_name (field):
  # get the column name of the field in df
  return app.schema[field]['field']


def read_summary ():
  """ read summary.csv """
  if hasattr(app, 'summary'):
    return app.summary

  fn = os.path.join(app.data_folder, 'summary.csv')
  smr = pd.read_csv(fn, na_filter=False)
  smr['uid'] = smr.apply(lambda r: r.name + 1, axis=1).astype(int)
  return smr


def read_results (field, dtype=str):
  """ read a result field """
  # read the result file
  info = app.schema[field]
  fn = os.path.join(app.data_folder, info['file'])
  df = pd.read_csv(fn, na_filter=False)
  col = info['field']
  return df[['uid', col]]


def read_results_batch (field_list):
  """ read multiple fields at once, minimizing file IO """
  fields = [app.schema[f] for f in field_list if f in app.schema]
  groups = group_by(fields, lambda x: x['file'])

  res = None
  for fn in groups:
    df = pd.read_csv(os.path.join(app.data_folder, fn), na_filter=False)
    names = ['uid'] + [d['name'] for d in groups[fn]]
    cols = ['uid'] + [d['field'] for d in groups[fn]]
    df = df[cols].rename(columns=dict(zip(cols, names)))
    res = df if res is None else pd.merge(res, df, on='uid')

  return res


def read_results_with_summary (field, dtype=str, diagnostics=True):
  """ read a result field and join with summary """
  # read results and join with summary
  smr = read_summary()
  results = read_results(field, dtype)
  col = app.schema[field]['field']
  df = pd.merge(smr, results, on='uid')

  # convert data type, remove Inf and NA
  n_df = df.shape[0]
  df = remove_na(df, col, dtype)

  # print warning messages
  if diagnostics:
    total = smr.shape[0]
    n_failed = total - n_df
    n_na = n_df - df.shape[0]
    if n_failed > 0 or n_na > 0:
      print_warn(f'Data quality warning: out of {total} universes, ')
      if n_failed > 0:
        percent = round(n_failed / total * 100, 1)
        print_warn(f' * {n_failed} universes ({percent}%) failed to run')
      if n_na > 0:
        percent = round(n_na / total * 100, 1)
        print_warn(f' * {n_na} {field} ({percent}%) contains Inf or NaN value')

  return df


def cluster_error (df):
  """ Cluster the error messages based on heuristics """
  if df.shape[0] < 1:
    df['group'] = pd.Series(dtype=str)
    return df

  # regex
  pt_skip = r'^(there were|warning message)'
  pt_err = r'^error'

  # row-wise function
  def process_row (row):
    sentences = row['message'].split('\n')
    first = ''
    i = 0
    while i < len(sentences):
      # skip the rows with uninformative message, and group by the first row
      if re.search(pt_skip, sentences[i], flags=re.IGNORECASE) is None:
        first = sentences[i]
        break
      i += 1
    
    # look for 'error' if the exit code is non-zero
    if row['exit_code'] > 0:
      while i < len(sentences): # start from where we left off
        if re.search(pt_err, sentences[i], flags=re.IGNORECASE) is not None:
          first = sentences[i]
          break
        i += 1

    return first

  df['group'] = df.apply(process_row, axis=1)
  return df


def sensitivity_f_test (df, col):
    """ Compute one-way F-test to estimate decision sensitivity """
    # compute one-way F-test
    res = {d['var']: sensitivity.sensitivity_f(df, d['var'], d['options'],
        col) for d in app.decisions}

    # check NaN
    for d in res:
        if np.isnan(res[d]):
            print_fail('ERROR: cannot compute sensitivity')
            print(f'F-test returns NaN value for decision "{d}"')
            exit(1)

    return res


def sensitivity_ks (df, col):
    """ compute Kolmogorov-Smirnov statistic """
    return {d['var']: sensitivity.sensitivity_ks(df, d['var'], d['options'],
        col) for d in app.decisions}


def sensitivity_ad (df, col):
    """ use k-samples Anderson-Darling test to compute sensitivity """
    return {d['var']: sensitivity.sensitivity_ad(df, d['var'], d['options'],
         col)[0] for d in app.decisions}


def cal_sensitivity(df=None):
    """ Compute sensitivity """
    # read the prediction and join with summary
    if df is None:
      df = read_results_with_summary('point_estimate', dtype=float)
    col = app.schema['point_estimate']['field']
    method = app.visualizer['sensitivity']

    if method == 'f':
        # one-way F-test
        score = sensitivity_f_test(df, col)
    if method == 'ks':
        # Kolmogorov-Smirnov statistic
        score = sensitivity_ks(df, col)
    if method == 'ad':
        # k-samples Anderson-Darling test
        score = sensitivity_ad(df, col)

    return score



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open("README.rst", "r") as fh:
    readme = fh.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['flask>=1.1.1', 'Click>=7.0', 'pandas>=1.0.1', 'scipy>=1.4.1',
    'boba>=1.1.1', 'flask-socketio>=5.0.0', 'apscheduler>=3.7.0',
    'scikit-learn>=0.24.1']

setup_requirements = []

test_requirements = []

setup(
    name='boba-visualizer',
    url='https://github.com/yyyliu/boba-visualizer',
    version='1.1.1',
    author="Yang Liu",
    author_email='yliu0@uw.edu',
    license="BSD license",
    description="Visualize multiverse outcomes",
    keywords='multiverse analysis',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    entry_points={
        'console_scripts': [
            'boba-server = bobaserver.run_server:main',
        ],
    },
    install_requires=requirements,
    long_description=readme + '\n\n' + history,
    packages=find_packages(include=['bobaserver', 'bobaserver.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    zip_safe=False,
    package_dir={'bobaserver': 'bobaserver/'},
    package_data={'bobaserver': ['./dist/*']},
    include_package_data=True
)



from flask import Flask
from flask_socketio import SocketIO
from apscheduler.schedulers.background import BackgroundScheduler

P_DIST = './dist/'

app = Flask(__name__, static_url_path='', static_folder=P_DIST)
socketio = SocketIO(app)
scheduler = BackgroundScheduler()

from bobaserver import routes
from bobaserver import monitor



import os
import numpy as np
import pandas as pd
import math
from flask import jsonify, request
from bobaserver import app
from .util import read_csv, read_json, read_key_safe, group_by, remove_na
import bobaserver.common as common


# entry
@app.route('/')
def index():
    return app.send_static_file('index.html')

# read the summary file
@app.route('/api/get_universes', methods=['POST'])
def get_universes():
    fn = os.path.join(app.data_folder, 'summary.csv')
    err, res = read_csv(fn, 0)
    reply = err if err else {'status': 'success', 'data': res[1:],
                             'header': res[0]}
    return jsonify(reply), 200

# read point estimates, p-value, fit metric value, and stacking weights
@app.route('/api/get_pred', methods=['POST'])
def get_pred():
    fields = ['point_estimate', 'p_value', 'fit', 'stacking_weight',
        'annotation', 'standard_error']
    res = common.read_results_batch(fields)
    header = res.columns.tolist()

    # remove Inf and NA in point estimates
    res = remove_na(res, 'point_estimate', dtype=float)

    res = [res[n].values.tolist() for n in header]
    reply = {'status': 'success', 'data': res, 'header': header,
        'sensitivity': app.sensitivity}
    return jsonify(reply), 200

# read uncertainty
@app.route('/api/get_uncertainty', methods=['POST'])
def get_uncertainty():
    f = app.schema['uncertainty']
    fn = os.path.join(app.data_folder, f['file'])
    err, res = read_csv(fn, 0)
    reply = err if err else {'status': 'success', 'data': res[1:]}
    if not err:
        header = ['uncertainty' if d == f['field'] else d for d in res[0]]
        reply['header'] = header

    return jsonify(reply), 200

# read the null distribution of point estimates
@app.route('/api/get_null', methods=['POST'])
def get_null():
    f = app.schema['null_distribution']
    fn = os.path.join(app.data_folder, f['file'])
    err, res = read_csv(fn, 0)
    reply = err if err else {'status': 'success', 'data': res[1:]}
    if not err:
        header = ['null_distribution' if d == f['field'] else d for d in res[0]]
        reply['header'] = header
    return jsonify(reply), 200

# read the overview, including decisions and ADG
@app.route('/api/get_overview', methods=['POST'])
def get_overview():
    res = {'schema': [app.schema[d]['name'] for d in app.schema],
        'decisions': app.decisions}
    res.update(app.visualizer)
    reply = {'status': 'success', 'data': res}
    return jsonify(reply), 200

# read the actual and predicted data of all data points in a universe
@app.route('/api/get_raw', methods=['POST'])
def get_raw():
    # fixme: prediction might not exist
    # fixme: now we assume specific column order, should use field name
    uid = request.json['uid']
    f = app.schema['prediction']
    fn = os.path.join(app.data_folder, f['file'].format(uid))
    err, res = read_csv(fn, 1)
    reply = err if err else {'status': 'success'}

    if not err:
        # sampling
        m = 100
        data = []
        for i in range(2):
            long = list(map(lambda d: float(d[i]), res))
            if len(long) > m:
                # quantile dot plot
                qt = np.append(np.arange(0, 1, 1 / m), 1.0)
                data.append(np.quantile(long, qt).tolist())
            else:
                data.append(long)

        # apply transform
        trans = read_key_safe(f, ['transform'], None)
        if trans:
            trans = trans.format('x')
            for i in range(2):
                data[i] = [eval(trans) for x in data[i]]

        reply['data'] = data

    return jsonify(reply), 200
